#pragma once
#include <stdexcept>
#include <mutex>
#include <vector>
#include <optional>
#include "app-window.h"
#include "rei-json/json.h"
#include "slint_string.h"
#include "slint.h"
#include "Anilist.h"
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <memory>
#include <ostream>
#include <string>
#include <print>

#define THREAD_COUNT 4

class AppWindow {

public:

	AppWindow(slint::ComponentHandle<ui::AppWindow> window) : ui(window) {
		listsSlint = std::make_shared<slint::VectorModel<ui::List>>();
		animesSlint = std::make_shared<slint::VectorModel<ui::Anime>>();

		for (int i = 0; i < THREAD_COUNT; ++i) {
			startThread(i);
		}

		loadDataFromDisk();
		initModels();
		setupCallbacks();
	}

	void startThread(int threadIndex) {
		waitAnimesLock[threadIndex].lock();
		thread_object[threadIndex] = std::thread([this, threadIndex]() {
			while (true) {
				waitAnimesLock[threadIndex].lock();
				shouldStopLoadingImages[threadIndex] = false;
				for (int i = threadIndex; i < animesSlint->row_count(); i += THREAD_COUNT) {
					continueAnimeLock[threadIndex].lock();
					animeListLock.lock();
					if (shouldStopLoadingImages[threadIndex]) {
						continueAnimeLock[threadIndex].unlock();
						animeListLock.unlock();
						break;
					}
					const int imageId = animesSlint->row_data(i)->id;
					const std::string imageUrl = std::string(animesSlint->row_data(i)->image_url);
					animeListLock.unlock();
					slint::Image image = getAnimeImage(imageId, imageUrl);
					slint::invoke_from_event_loop([this, image, threadIndex, i] { 
						animeListLock.lock();
						if (shouldStopLoadingImages[threadIndex]) {
							continueAnimeLock[threadIndex].unlock();
							animeListLock.unlock();
							return;
						}
						auto row = animesSlint->row_data(i);
						if (!row.has_value()) {
							std::println(std::cerr, "Can't find row data index {}", i);
							return;
						}
						row->image = image;
						animesSlint->set_row_data(i, row.value());
						animeListLock.unlock();
						continueAnimeLock[threadIndex].unlock();
					});
				}
				std::println("thread {} finished loading images", threadIndex);
			}
		});
	}

	void run() {
		ui->run();
	}

	void loadDataFromDisk() {
		std::filesystem::create_directories("./imgs");

		std::ifstream file("./save.json");
		if (!file.is_open()) {
			std::println(std::cerr, "can't open file");
			json = rei::json::JsonObject{};
			json.set("version", 1);
			json.addArray("lists", rei::json::JsonArray{});
			return;
		}
		std::stringstream buffer;
		buffer << file.rdbuf();
		auto jsonStr = buffer.str();
		json = std::get<rei::json::JsonObject>(rei::json::parse(jsonStr));
	}

	void initModels() {

		listsSlint = std::make_shared<slint::VectorModel<ui::List>>();
		ui->global<ui::State>().set_lists(listsSlint);

		animesSlint = std::make_shared<slint::VectorModel<ui::Anime>>();
		ui->global<ui::State>().set_current_list({
			.name = "ok",
			.animes = animesSlint
		});

		auto& lists = json.getArray("lists");

		lists.forEach([&](rei::json::FieldValue& element, int index) {
				auto& obj = element.asObject();
				listsSlint->push_back(ui::List{
					.name = slint::SharedString(obj.getString("name")),
					.selected = false
				});
		});
		if (lists.count() > 0) {
			selectList(0);
		}
	}

	void setupCallbacks() {
		ui->global<ui::State>().on_select_list([&](int index) {
			selectList(index);
		});

		ui->global<ui::State>().on_add_anilist_list([&](ui::AddAnilistListParams params) {
			addAnilistList(std::string(params.name), std::string(params.anilist_user_name), std::string(params.anilist_list_name));
		});

		ui->global<ui::State>().on_sync_list([&](slint::SharedString name) {
			//std::println("{}",res->body);
			try {
				//addAnilistList("defaut", name, "Completed");
				

				/*std::println("0000");*/
				/*selectList(saveJson.getArray("lists").getObject(0).getArray("animes"), animesSlint);*/
				/*std::println("1111");*/
				/*imgsToLoad.clear();*/
				/*std::println("2222");*/
				/*saveJson.getArray("lists").getObject(0).getArray("animes").forEach([&] (rei::json::FieldValue& element, int index) {*/
					/*std::println("IN");*/
					/*imgsToLoad.push_back(element.asObject().getObject("coverImage").getString("extraLarge"));*/
					/*std::println("OUT");*/
				/*});*/
				/*std::println("3333");*/
				/*readLinks.unlock();*/

			} catch (std::exception& e) {
				std::println(std::cerr, "Error: {}", e.what());
			} catch (char const* e) {
				std::println(std::cerr, "Error str: {}", e);
			}

		});
	}

	void selectList(int listIndex) {
		auto& list = json.getArray("lists").getObject(listIndex);
		auto& animes = list.getArray("animes");

		for (int i = 0; i < listsSlint->row_count(); ++i) {
			auto row = listsSlint->row_data(i);
			if (row->selected) {
				row->selected = false;
				listsSlint->set_row_data(i, row.value());
			} else if (i == listIndex) {
				row->selected = true;
				listsSlint->set_row_data(i, row.value());
			}
		}

		animeListLock.lock();
		for (int i = 0; i < THREAD_COUNT; ++i) {
			shouldStopLoadingImages[i] = true;
		}
		animesSlint->clear();
		animes.forEach([&](rei::json::FieldValue& element, int index) {
			auto& obj = element.asObject();
			animesSlint->push_back(ui::Anime{
				.id = obj.getNumber("id"),
				.title = slint::SharedString(obj.getObject("title").getString("romaji")),
				.description = slint::SharedString(obj.getString("description")),
				.image_url = slint::SharedString(obj.getObject("coverImage").getString("extraLarge"))
				//.image = getAnimeImage(fakeAnimeId, obj.getString("coverImageUrl"))
			});
		});
		animeListLock.unlock();
		for (int i = 0; i < THREAD_COUNT; ++i) {
			waitAnimesLock[i].unlock();
		}

		ui->global<ui::State>().set_current_list({
			.name = slint::SharedString(list.getString("name")),
			.animes = animesSlint
		});
		
	}

	void addAnilistList(const std::string& listName, const std::string& anilistUserName, const std::string& anilistListName) {
		std::println("Fetching user list {}->{}", anilistUserName, anilistListName);
		auto anilistJsonResult = Anilist::fetchUserLists(anilistUserName);
		if (!anilistJsonResult.has_value()) {
			return;
		}
		auto anilistJson = anilistJsonResult.value();

		rei::json::JsonObject saveJson{};
		saveJson.set("version", 1);

		rei::json::JsonObject listJson{};
		anilistJson.getObject("data").getObject("MediaListCollection").getArray("lists")
			.forEach([&](rei::json::FieldValue& field, int index) {
				if (field.asObject().getString("name") != anilistListName)	 {
					return ;
				}
				listJson.set("name", listName);
				listJson.set("source", "anilist");
				listJson.set("anilistUserName", anilistUserName);
				listJson.set("anilistListName", anilistListName);

				rei::json::JsonArray animesJson{};
				field.asObject().getArray("entries").forEach([&](rei::json::FieldValue& value, int index) {
						auto animeJson = value.asObject().getObject("media");
						animesJson.pushObject(animeJson);
				});

				listJson.addArray("animes", animesJson);
			});
		json.getArray("lists").pushObject(listJson);
		listsSlint->push_back(ui::List{
			.name = slint::SharedString(listName),
			.selected = false
		});
		save();
	}

	slint::Image getAnimeImage(int animeId, std::string imageUrl) {
		//std::println("Get image for anime {}", animeId);
		std::string imagePath = "./imgs/" + std::to_string(animeId);

		if (imageUrl.ends_with(".jpg") || imageUrl.ends_with(".jpeg")) {
			imagePath += ".jpg";
		} else if (imageUrl.ends_with(".png")) {
			imagePath += ".png";
		}

		if (std::filesystem::exists(imagePath)) {
			//std::println("Image already exists {}", animeId);
			auto image = slint::Image::load_from_path(imagePath.c_str());
			//std::println("Image loaded {}", animeId);
			return image;
		}

		//std::println("Image doesn't exist, fetching {}", animeId);

		
		std::erase(imageUrl, '\\');
		const auto imageData = Anilist::fetchImage(imageUrl);

		//std::println("Image fetched {}", animeId);

		if (!imageData.has_value()) {
			//std::println(std::cerr, "Cannot fetch image {}", imageUrl);
			throw "Cannot fetch image";
		}

		std::ofstream imageFile(imagePath);
		if (!imageFile.is_open()) {
			throw "Can't save image";
		}

		imageFile << imageData.value();
		imageFile.close();

		//std::println("Image saved {}", animeId);

		//std::println("Image loaded {}", animeId);
		return slint::Image::load_from_path(imagePath.c_str());
	}

	void save() {
		std::ofstream save("./save.json");
		if (!save.is_open()) {
			throw std::runtime_error("Can't save data");
		}
		save << rei::json::toString(json);
		save.close();
	}

private:

	rei::json::JsonObject json;
	slint::ComponentHandle<ui::AppWindow> ui;
	std::shared_ptr<slint::VectorModel<ui::Anime>> animesSlint;
	std::shared_ptr<slint::VectorModel<ui::List>> listsSlint;
	std::mutex animeListLock;
	std::mutex waitAnimesLock[THREAD_COUNT];
	std::mutex continueAnimeLock[THREAD_COUNT];
	bool shouldStopLoadingImages[THREAD_COUNT] = {false};
	std::thread thread_object[THREAD_COUNT];
};
